/**
 * Z-Wave Certified Application Switch On/Off
 *
 * @copyright 2018 Silicon Laboratories Inc.
 */

/****************************************************************************/
/*                              INCLUDE FILES                               */
/****************************************************************************/
#include "config_rf.h"
#include <stdbool.h>
#include <stdint.h>
#include "SizeOf.h"
#include "Assert.h"
#include <SWO_Debug.h>
#include "DebugPrintConfig.h"
#define DEBUGPRINT
#include "DebugPrint.h"
#include "config_app.h"
#include "ZAF_app_version.h"
#include <ZAF_file_ids.h>
#include "nvm3.h"
#include "ZAF_nvm3_app.h"
#include <em_system.h>
#include <ZW_radio_api.h>
#include <ZW_slave_api.h>
#include <ZW_classcmd.h>
#include <ZW_TransportLayer.h>

#include <ZAF_uart_utils.h>

#include <misc.h>
#include <string.h>
#include <ev_man.h>
/*IO control*/
#include <board.h>

#include <AppTimer.h>
#include <SwTimer.h>
#include <EventDistributor.h>
#include <ZW_system_startup_api.h>
#include <ZW_application_transport_interface.h>

#include <association_plus.h>
#include <agi.h>
#include <CC_Association.h>
#include <CC_AssociationGroupInfo.h>
#include <CC_Basic.h>
#include <CC_BinarySwitch.h>
#include <CC_DeviceResetLocally.h>
#include <CC_Indicator.h>
#include <CC_MultiChan.h>
#include <CC_MultiChanAssociation.h>
#include <CC_Notification.h>
#include <CC_PowerLevel.h>
#include <CC_Supervision.h>
#include <CC_Version.h>
#include <CC_ZWavePlusInfo.h>
#include <CC_FirmwareUpdate.h>
#include <CC_ManufacturerSpecific.h>
#include <notification.h>
#include <zaf_event_helper.h>
#include <zaf_job_helper.h>

#include <ZAF_Common_helper.h>
#include <ZAF_network_learn.h>
#include <ota_util.h>
#include <ZAF_TSE.h>
#include <ZAF_CmdPublisher.h>
#include <em_wdog.h>
/****************************************************************************/
/* Application specific button and LED definitions                          */
/****************************************************************************/

#define SWITCH_TOGGLE_BTN   APP_BUTTON_A
#define SWITCH_TOGGLE_BTN2  APP_BUTTON_B
#define SWITCH_STATUS_LED   APP_LED_A

/* Ensure we did not allocate the same physical button or led to more than one function */
STATIC_ASSERT(APP_BUTTON_LEARN_RESET != SWITCH_TOGGLE_BTN,
              STATIC_ASSERT_FAILED_button_overlap);
STATIC_ASSERT(APP_LED_INDICATOR != SWITCH_STATUS_LED,
              STATIC_ASSERT_FAILED_led_overlap);

/****************************************************************************/
/*                      PRIVATE TYPES and DEFINITIONS                       */
/****************************************************************************/

/**
 * Application events for AppStateManager(..)
 */
typedef enum _EVENT_APP_
{
	EVENT_EMPTY = DEFINE_EVENT_APP_NBR,
	  EVENT_APP_INIT,
	  EVENT_APP_REFRESH_MMI,
	  EVENT_APP_FLUSHMEM_READY,
	  EVENT_APP_NEXT_EVENT_JOB,
	  EVENT_APP_FINISH_EVENT_JOB,
	  EVENT_APP_GET_NODELIST,
	  EVENT_APP_SEND_OVERLOAD_NOTIFICATION,
	  EVENT_APP_SMARTSTART_IN_PROGRESS
} EVENT_APP;


/**
 * Application states. Function AppStateManager(..) includes the state
 * event machine.
 */
typedef enum _STATE_APP_
{
  STATE_APP_STARTUP,
  STATE_APP_IDLE,
  STATE_APP_LEARN_MODE,
  STATE_APP_RESET,
  STATE_APP_TRANSMIT_DATA
} STATE_APP;

#define BIN_SWITCH_1        (ENDPOINT_1 - 1)
#define BIN_SWITCH_2        (ENDPOINT_2 - 1)

uint8_t supportedEvents = NOTIFICATION_EVENT_POWER_MANAGEMENT_OVERLOADED_DETECTED;

/****************************************************************************/
/*                              PRIVATE DATA                                */
/****************************************************************************/

/**
 * Node information list for nodes included Non-Securely.
 * The command classes in this list are supported when the node
 * is included without security S0/S2. They are also listed in the NIF
 * that is sent when the node is excluded.
 * CHANGE THIS - Add all supported non-secure command classes here
 **/
static uint8_t cmdClassListNonSecureNotIncluded[] =
{
  COMMAND_CLASS_ZWAVEPLUS_INFO,
  COMMAND_CLASS_SWITCH_BINARY_V2,
  COMMAND_CLASS_ASSOCIATION,
  COMMAND_CLASS_MULTI_CHANNEL_ASSOCIATION_V2,
  COMMAND_CLASS_ASSOCIATION_GRP_INFO,
  COMMAND_CLASS_NOTIFICATION_V3,
  COMMAND_CLASS_TRANSPORT_SERVICE_V2,
  COMMAND_CLASS_VERSION,
  COMMAND_CLASS_MANUFACTURER_SPECIFIC,
  COMMAND_CLASS_DEVICE_RESET_LOCALLY,
  COMMAND_CLASS_INDICATOR,
  COMMAND_CLASS_POWERLEVEL,
  COMMAND_CLASS_SECURITY,
  COMMAND_CLASS_SECURITY_2,
  COMMAND_CLASS_MULTI_CHANNEL_V4,
  COMMAND_CLASS_SUPERVISION,
  COMMAND_CLASS_FIRMWARE_UPDATE_MD_V5
};


/**
 * Partial Node information list for nodes included securely.
 * The command classes that are supported non-securely when the
 * node is included with security protocols S0/S2.
 * Be sure Command classes in this list are NOT duplicated in the
 * list cmdClassListSecure[] below.
 * CHANGE THIS - Add all supported non-secure command classes here
 **/
static uint8_t cmdClassListNonSecureIncludedSecure[] =
{
  COMMAND_CLASS_ZWAVEPLUS_INFO,
  COMMAND_CLASS_TRANSPORT_SERVICE_V2,
  COMMAND_CLASS_SECURITY,
  COMMAND_CLASS_SECURITY_2,
  COMMAND_CLASS_SUPERVISION
};

/**
 * Partial Node information list for nodes included securely.
 * The command classes that are supported only when the
 * node is included with security protocols S0/S2.
 * Be sure Command classes in this list are NOT duplicated in the
 * list cmdClassListNonSecureIncludedSecure[] above.
 * CHANGE THIS - Add all supported secure command classes here
 **/
static uint8_t cmdClassListSecure[] =
{
		COMMAND_CLASS_SWITCH_BINARY_V2,
		COMMAND_CLASS_ASSOCIATION,
		COMMAND_CLASS_MULTI_CHANNEL_ASSOCIATION_V2,
		COMMAND_CLASS_ASSOCIATION_GRP_INFO,
		COMMAND_CLASS_NOTIFICATION_V3,
		COMMAND_CLASS_VERSION,
		COMMAND_CLASS_MANUFACTURER_SPECIFIC,
		COMMAND_CLASS_DEVICE_RESET_LOCALLY,
		COMMAND_CLASS_INDICATOR,
		COMMAND_CLASS_POWERLEVEL,
		COMMAND_CLASS_MULTI_CHANNEL_V4,
		COMMAND_CLASS_FIRMWARE_UPDATE_MD_V5
};

/**
 * Structure includes application node information list's and device type.
 */
APP_NODE_INFORMATION m_AppNIF =
{
  cmdClassListNonSecureNotIncluded, sizeof(cmdClassListNonSecureNotIncluded),
  cmdClassListNonSecureIncludedSecure, sizeof(cmdClassListNonSecureIncludedSecure),
  cmdClassListSecure, sizeof(cmdClassListSecure),
  DEVICE_OPTIONS_MASK, {GENERIC_TYPE, SPECIFIC_TYPE}
};

/**
* Set up security keys to request when joining a network.
* These are taken from the config_app.h header file.
*/
static const uint8_t SecureKeysRequested = REQUESTED_SECURITY_KEYS;

static const SAppNodeInfo_t AppNodeInfo =
{
  .DeviceOptionsMask = DEVICE_OPTIONS_MASK,
  .NodeType.generic = GENERIC_TYPE,
  .NodeType.specific = SPECIFIC_TYPE,
  .CommandClasses.UnSecureIncludedCC.iListLength = sizeof_array(cmdClassListNonSecureNotIncluded),
  .CommandClasses.UnSecureIncludedCC.pCommandClasses = cmdClassListNonSecureNotIncluded,
  .CommandClasses.SecureIncludedUnSecureCC.iListLength = sizeof_array(cmdClassListNonSecureIncludedSecure),
  .CommandClasses.SecureIncludedUnSecureCC.pCommandClasses = cmdClassListNonSecureIncludedSecure,
  .CommandClasses.SecureIncludedSecureCC.iListLength = sizeof_array(cmdClassListSecure),
  .CommandClasses.SecureIncludedSecureCC.pCommandClasses = cmdClassListSecure
};

static const SRadioConfig_t RadioConfig =
{
  .iListenBeforeTalkThreshold = ELISTENBEFORETALKTRESHOLD_DEFAULT,
  .iTxPowerLevelMax = APP_MAX_TX_POWER,
  .iTxPowerLevelAdjust = APP_MEASURED_0DBM_TX_POWER,
  .eRegion = REGION_EU
};

static const SProtocolConfig_t ProtocolConfig = {
  .pVirtualSlaveNodeInfoTable = NULL,
  .pSecureKeysRequested = &SecureKeysRequested,
  .pNodeInfo = &AppNodeInfo,
  .pRadioConfig = &RadioConfig
};


static uint8_t ep12_noSec_InclNonSecure[] =
{
  COMMAND_CLASS_ZWAVEPLUS_INFO,
  COMMAND_CLASS_SWITCH_BINARY_V2,
  COMMAND_CLASS_ASSOCIATION,
  COMMAND_CLASS_ASSOCIATION_GRP_INFO,
  COMMAND_CLASS_MULTI_CHANNEL_ASSOCIATION_V2,
  COMMAND_CLASS_NOTIFICATION_V3,
  COMMAND_CLASS_SUPERVISION
};


static uint8_t ep12_noSec_InclSecure[] =
{
  COMMAND_CLASS_ZWAVEPLUS_INFO,
  COMMAND_CLASS_SUPERVISION,
  COMMAND_CLASS_SECURITY,
  COMMAND_CLASS_SECURITY_2
};

static uint8_t ep12_sec_InclSecure[] =
{
  COMMAND_CLASS_SWITCH_BINARY_V2,
  COMMAND_CLASS_ASSOCIATION,
  COMMAND_CLASS_ASSOCIATION_GRP_INFO,
  COMMAND_CLASS_MULTI_CHANNEL_ASSOCIATION_V2,
  COMMAND_CLASS_NOTIFICATION_V3
};


static uint8_t ep3_noSec_InclNonSecure[] =
{
  COMMAND_CLASS_ZWAVEPLUS_INFO,
  COMMAND_CLASS_SWITCH_BINARY_V2,
  COMMAND_CLASS_ASSOCIATION,
  COMMAND_CLASS_ASSOCIATION_GRP_INFO,
  COMMAND_CLASS_MULTI_CHANNEL_ASSOCIATION_V2,
  COMMAND_CLASS_NOTIFICATION_V3,
  COMMAND_CLASS_SUPERVISION
};


static uint8_t ep3_noSec_InclSecure[] =
{
  COMMAND_CLASS_ZWAVEPLUS_INFO,
  COMMAND_CLASS_SUPERVISION,
  COMMAND_CLASS_SECURITY,
  COMMAND_CLASS_SECURITY_2
};

static uint8_t ep3_sec_InclSecure[] =
{
  COMMAND_CLASS_SWITCH_BINARY_V2,
  COMMAND_CLASS_ASSOCIATION,
  COMMAND_CLASS_ASSOCIATION_GRP_INFO,
  COMMAND_CLASS_MULTI_CHANNEL_ASSOCIATION_V2,
  COMMAND_CLASS_NOTIFICATION_V3
};

static uint8_t sessionID;

static EP_NIF endpointsNIF[NUMBER_OF_ENDPOINTS] =
{
 /*
  * Endpoint 1
  */
  {
    GENERIC_TYPE_SWITCH_BINARY,
    SPECIFIC_TYPE_NOT_USED,
    {
      {ep12_noSec_InclNonSecure, sizeof(ep12_noSec_InclNonSecure)},
      {{ep12_noSec_InclSecure, sizeof(ep12_noSec_InclSecure)}, {ep12_sec_InclSecure, sizeof(ep12_sec_InclSecure)}}
    }
  },
  /*
   * Endpoint 2
   */
  {
    GENERIC_TYPE_SWITCH_BINARY,
    SPECIFIC_TYPE_NOT_USED,
    {
    		{ep12_noSec_InclNonSecure, sizeof(ep12_noSec_InclNonSecure)},
    		{{ep12_noSec_InclSecure, sizeof(ep12_noSec_InclSecure)}, {ep12_sec_InclSecure, sizeof(ep12_sec_InclSecure)}}
    }
  }
};

EP_FUNCTIONALITY_DATA endPointFunctionality =
{
  {
    NUMBER_OF_INDIVIDUAL_ENDPOINTS,     /**< nbrIndividualEndpoints 7 bit*/
    RES_ZERO,                           /**< resIndZeorBit 1 bit*/
    NUMBER_OF_AGGREGATED_ENDPOINTS,     /**< nbrAggregatedEndpoints 7 bit*/
    RES_ZERO,                           /**< resAggZeorBit 1 bit*/
    RES_ZERO,                           /**< resZero 6 bit*/
    ENDPOINT_IDENTICAL_DEVICE_CLASS_NO,/**< identical 1 bit*/
    ENDPOINT_DYNAMIC_NO                /**< dynamic 1 bit*/
  }
};

/**
 * Setup AGI lifeline table from config_app.h
 */
cc_group_t  agiTableLifeLine[] = {AGITABLE_LIFELINE_GROUP};

static const AGI_PROFILE lifelineProfile =
{
 ASSOCIATION_GROUP_INFO_REPORT_PROFILE_GENERAL_V2,
 ASSOCIATION_GROUP_INFO_REPORT_PROFILE_GENERAL_LIFELINE_V2
};

cc_group_t agiTableLifeLineEP1[] = {AGITABLE_LIFELINE_GROUP_EP1};
cc_group_t agiTableLifeLineEP2[] = {AGITABLE_LIFELINE_GROUP_EP2};

/**
 * Setup AGI root device groups table from config_app.h
 */
const AGI_GROUP agiTableRootDeviceGroups[] = {AGITABLE_ROOTDEVICE_GROUPS};
const AGI_GROUP agiTableEndpoint1Groups[] = {AGITABLE_ENDPOINT_1_GROUPS};
const AGI_GROUP agiTableEndpoint2Groups[] = {AGITABLE_ENDPOINT_2_GROUPS};

#define LIFELANE_PROFILE  (&agiTableRootDeviceGroups[notificationOverLoadendpoint - 1].profile)
#define LIFELANE_GROUP_PROFILE  (&(agiTableEndpoint1Groups->profile))

ST_ENDPOINT_ICONS ZWavePlusEndpointIcons[] = {ENDPOINT_ICONS};
/**
 * Application state-machine state.
 */
static STATE_APP currentState = STATE_APP_IDLE;
static RECEIVE_OPTIONS_TYPE_EX rxOptionSupervision;
/**
 * Parameter is used to save wakeup reason after ApplicationInit(..)
 */
EResetReason_t g_eResetReason;

#define SWITCH_ON           (0x01)
#define SWITCH_OFF          (0x00)

typedef struct _BIN_SWITCH_{
  uint8_t switchStatus;
}BIN_SWITCH;

BIN_SWITCH BIN_SWITCH1;
BIN_SWITCH BIN_SWITCH2;

ASSOCIATION_ROOT_GROUP_MAPPING rootGroupMapping[] = { ASSOCIATION_ROOT_GROUP_MAPPING_CONFIG};

static ESwTimerStatus notificationOverLoadTimerStatus = ESWTIMER_STATUS_FAILED;
// Timer
static SSwTimer NotificationTimer;
static SSwTimer SupervisionTimer;

static bool notificationOverLoadActiveState =  false;
static uint8_t notificationOverLoadendpoint = 0;


/**
 * LED state
 */
static CMD_CLASS_BIN_SW_VAL onOffState = CMD_CLASS_BIN_OFF;

static TaskHandle_t g_AppTaskHandle;

#ifdef DEBUGPRINT
static uint8_t m_aDebugPrintBuffer[96];
#endif

// Pointer to AppHandles that is passed as input to ApplicationTask(..)
SApplicationHandles* g_pAppHandles;

// Prioritized events that can wakeup protocol thread.
typedef enum EApplicationEvent
{
  EAPPLICATIONEVENT_TIMER = 0,
  EAPPLICATIONEVENT_ZWRX,
  EAPPLICATIONEVENT_ZWCOMMANDSTATUS,
  EAPPLICATIONEVENT_APP
} EApplicationEvent;

static void EventHandlerZwRx(void);
static void EventHandlerZwCommandStatus(void);
static void EventHandlerApp(void);

// Event distributor object
static SEventDistributor g_EventDistributor;

// Event distributor event handler table
static const EventDistributorEventHandler g_aEventHandlerTable[4] =
{
  AppTimerNotificationHandler,  // Event 0
  EventHandlerZwRx,
  EventHandlerZwCommandStatus,
  EventHandlerApp
};

// Used by the application data file.
typedef struct SApplicationData
{
  uint8_t onOffState;
} SApplicationData;

#define FILE_SIZE_APPLICATIONDATA     (sizeof(SApplicationData))


// FileSet - allows FileSystemVerifySet to validate required files
static const SFileDescriptor g_aFileDescriptors[] = {
{ .ObjectKey = ZAF_FILE_ID_ASSOCIATIONINFO,   .iDataSize = ZAF_FILE_SIZE_ASSOCIATIONINFO },
{ .ObjectKey = FILE_ID_APPLICATIONDATA,       .iDataSize = FILE_SIZE_APPLICATIONDATA     },
{ .ObjectKey = ZAF_FILE_ID_APP_VERSION,       .iDataSize = ZAF_FILE_SIZE_APP_VERSION     }
};

static EFileSysVerifyStatus g_aFileStatus[sizeof_array(g_aFileDescriptors)];

static SApplicationData ApplicationData;

#define APP_EVENT_QUEUE_SIZE 5

/**
 * The following four variables are used for the application event queue.
 */
static SQueueNotifying m_AppEventNotifyingQueue;
static StaticQueue_t m_AppEventQueueObject;
static EVENT_APP eventQueueStorage[APP_EVENT_QUEUE_SIZE];
static QueueHandle_t m_AppEventQueue;



/* Indicate here which End Points (NOT including root device) support each command class */
static const uint8_t endpoints_Supporting_BinarySwitch_CC[] = {1,2};
s_CC_binarySwitch_data_t ZAF_TSE_BinarySwitchData[sizeof_array(endpoints_Supporting_BinarySwitch_CC)];
s_CC_indicator_data_t ZAF_TSE_localActuationIdentifyData = {
  .rxOptions = {
    .rxStatus = 0,          /* rxStatus, verified by the TSE for Multicast */
    .securityKey = 0,       /* securityKey, ignored by the TSE */
    .sourceNode = {0,0},    /* sourceNode (nodeId, endpoint), verified against lifeline destinations by the TSE */
    .destNode = {0,0}       /* destNode (nodeId, endpoint), verified by the TSE for local endpoint */
  },
  .indicatorId = 0x50      /* Identify Indicator*/
};





/**
* TSE simulated RX option for local change addressed to the Root Device
* All applications can use this variable when triggering the TSE after
* a local / non Z-Wave initiated change
*/
static RECEIVE_OPTIONS_TYPE_EX zaf_tse_local_actuation = {
    0,      /* rxStatus, verified by the TSE for Multicast */
    0,      /* securityKey, ignored by the TSE */
    {0,0},  /* sourceNode (nodeId, endpoint), verified against lifeline destinations by the TSE */
    {0,1}   /* destNode (nodeId, endpoint), verified by the TSE for local endpoint */
};

static RECEIVE_OPTIONS_TYPE_EX zaf_tse_local_actuation2 = {
    0,      /* rxStatus, verified by the TSE for Multicast */
    0,      /* securityKey, ignored by the TSE */
    {0,0},  /* sourceNode (nodeId, endpoint), verified against lifeline destinations by the TSE */
    {0,2}   /* destNode (nodeId, endpoint), verified by the TSE for local endpoint */
};

static RECEIVE_OPTIONS_TYPE_EX zaf_tse_local_ep1_actuation = {
    .rxStatus = 0,        /* rxStatus, verified by the TSE for Multicast */
    .securityKey = 0,     /* securityKey, ignored by the TSE */
    .sourceNode = {0,0},  /* sourceNode (nodeId, endpoint), verified against lifeline destinations by the TSE */
    .destNode = {0,1}     /* destNode (nodeId, endpoint), verified by the TSE for local endpoint */
};

/* TSE simulated RX option for local change addressed to End Point 2 */
static RECEIVE_OPTIONS_TYPE_EX zaf_tse_local_ep2_actuation = {
    .rxStatus = 0,        /* rxStatus, verified by the TSE for Multicast */
    .securityKey = 0,     /* securityKey, ignored by the TSE */
    .sourceNode = {0,0},  /* sourceNode (nodeId, endpoint), verified against lifeline destinations by the TSE */
    .destNode = {0,2}     /* destNode (nodeId, endpoint), verified by the TSE for local endpoint */
};



static nvm3_Handle_t* pFileSystemApplication;

/****************************************************************************/
/*                              EXPORTED DATA                               */
/****************************************************************************/

/****************************************************************************/
/*                            PRIVATE FUNCTIONS                             */
/****************************************************************************/
void DeviceResetLocallyDone(TRANSMISSION_RESULT * pTransmissionResult);
void DeviceResetLocally(void);
STATE_APP GetAppState(void);
void AppStateManager(EVENT_APP event);
static void ChangeState(STATE_APP newState);

uint8_t CC_BinarySwitch_getEndpointIndex(uint8_t endpoint);


static void ApplicationTask(SApplicationHandles* pAppHandles);

bool LoadConfiguration(void);
void SetDefaultConfiguration(void);

static void LearnCompleted(void);

void ToggleLed(void);
void RefreshMMI(void);
SApplicationData readAppData(void);
void writeAppData(const SApplicationData* pAppData);

void AppResetNvm(void);

static void LearnCompleted(void);

void ZCB_JobStatus(TRANSMISSION_RESULT * pTransmissionResult);
void ZCB_NotificationTimerCallback(SSwTimer *pTimer);
static void ToggleSwitch(uint8_t switchID);
static void notificationToggle(void);
//void ZCB_CommandClassSupervisionGetReceived(SUPERVISION_GET_RECEIVED_HANDLER_ARGS * pArgs);
void ZCB_SupervisionTimerCallback(SSwTimer *pTimer);





/**
* @brief Called when protocol puts a frame on the ZwRxQueue.
*/
static void EventHandlerZwRx(void)
{
  QueueHandle_t Queue = g_pAppHandles->ZwRxQueue;
  SZwaveReceivePackage RxPackage;

  // Handle incoming replies
  while (xQueueReceive(Queue, (uint8_t*)(&RxPackage), 0) == pdTRUE)
  {
    DPRINT("Incoming Rx msg\r\n");

    switch (RxPackage.eReceiveType)
    {
      case EZWAVERECEIVETYPE_SINGLE:
      {
        ZAF_CP_CommandPublish(ZAF_getCPHandle(), &RxPackage);
        break;
      }

      case EZWAVERECEIVETYPE_NODE_UPDATE:
      {
        /*ApplicationSlaveUpdate() was called from this place in version prior to SDK7*/
        break;
      }

      case EZWAVERECEIVETYPE_SECURITY_EVENT:
      {
        /*ApplicationSecurityEvent() was used to support CSA, not needed in SDK7*/
        break;
      }

    default:
      {
        ASSERT(false);
        break;
      }
    }
  }

}


/**
* @brief Triggered when protocol puts a message on the ZwCommandStatusQueue.
*/
static void EventHandlerZwCommandStatus(void)
{
  QueueHandle_t Queue = g_pAppHandles->ZwCommandStatusQueue;
  SZwaveCommandStatusPackage Status;

  // Handle incoming replies
  while (xQueueReceive(Queue, (uint8_t*)(&Status), 0) == pdTRUE)
  {
    DPRINT("Incoming Status msg\r\n");

    switch (Status.eStatusType)
    {
      case EZWAVECOMMANDSTATUS_TX:
      {
        SZWaveTransmitStatus* pTxStatus = &Status.Content.TxStatus;
        if (!pTxStatus->bIsTxFrameLegal)
        {
          DPRINT("Auch - not sure what to do\r\n");
        }
        else
        {
          DPRINT("Tx Status received\r\n");
          if (pTxStatus->Handle)
          {
            void(*pCallback)(uint8_t txStatus, TX_STATUS_TYPE* extendedTxStatus) = pTxStatus->Handle;
            pCallback(pTxStatus->TxStatus, &pTxStatus->ExtendedTxStatus);
          }
        }

        break;
      }

      case EZWAVECOMMANDSTATUS_GENERATE_RANDOM:
      {
        DPRINT("Generate Random status\r\n");
        break;
      }

      case EZWAVECOMMANDSTATUS_LEARN_MODE_STATUS:
      {
        DPRINTF("Learn status %d\r\n", Status.Content.LearnModeStatus.Status);
        if (ASSIGN_COMPLETE == Status.Content.LearnModeStatus.Status)
        {
          LearnCompleted();
        }
        else if(APPLICATION_NETWORK_LEARN_MODE_COMPLETED_TIMEOUT == Status.Content.LearnModeStatus.Status)
        {
          ZAF_EventHelperEventEnqueue((EVENT_APP) EVENT_SYSTEM_LEARNMODE_FINISHED);
        }
        else if(APPLICATION_NETWORK_LEARN_MODE_COMPLETED_FAILED == Status.Content.LearnModeStatus.Status)
        {
          //Reformats protocol and application NVM. Then soft reset.
          ZAF_EventHelperEventEnqueue((EVENT_APP) EVENT_SYSTEM_RESET);
        }
        break;
      }

      case EZWAVECOMMANDSTATUS_NETWORK_LEARN_MODE_START:
      {
        break;
      }

      case EZWAVECOMMANDSTATUS_SET_DEFAULT:
      { // Received when protocol is started (not implemented yet), and when SetDefault command is completed
        DPRINTF("Protocol Ready\r\n");
        ZAF_EventHelperEventEnqueue(EVENT_APP_FLUSHMEM_READY);
        break;
      }

      case EZWAVECOMMANDSTATUS_INVALID_TX_REQUEST:
      {
        DPRINTF("ERROR: Invalid TX Request to protocol - %d", Status.Content.InvalidTxRequestStatus.InvalidTxRequest);
        break;
      }

      case EZWAVECOMMANDSTATUS_INVALID_COMMAND:
      {
        DPRINTF("ERROR: Invalid command to protocol - %d", Status.Content.InvalidCommandStatus.InvalidCommand);
        break;
      }

      default:
        ASSERT(false);
        break;
    }
  }
}

void ZCB_CommandClassSupervisionGetReceived(SUPERVISION_GET_RECEIVED_HANDLER_ARGS * pArgs)
{
  if (SWITCH_MULTILEVEL_SET == pArgs->cmd && COMMAND_CLASS_SWITCH_MULTILEVEL_V4 == pArgs->cmdClass)
  {
    uint32_t duration = 0;
    pArgs->duration = duration;
    if (0 < duration)
    {
      pArgs->status = CC_SUPERVISION_STATUS_WORKING;

      if (CC_SUPERVISION_STATUS_UPDATES_SUPPORTED == CC_SUPERVISION_EXTRACT_STATUS_UPDATE(pArgs->properties1))
      {
        if (TimerIsActive(&SupervisionTimer))
        {
          // The timer is active. It means we have a WORKING session pending from previous Supervision Get command.
          // Complete that session first. By stopping the timer and manually calling the timer callback function.
          TimerStop(&SupervisionTimer);
          ZCB_SupervisionTimerCallback(NULL);
        }

        // Save the data
        rxOptionSupervision = *(pArgs->rxOpt);
        sessionID = CC_SUPERVISION_EXTRACT_SESSION_ID(pArgs->properties1);

        pArgs->properties1 = CC_SUPERVISION_ADD_MORE_STATUS_UPDATE(CC_SUPERVISION_MORE_STATUS_UPDATES_REPORTS_TO_FOLLOW) | CC_SUPERVISION_ADD_SESSION_ID(sessionID);

        /*
         * Start timer that will send another Supervision report when triggered.
         *
         * Checking whether duration is higher than 127 is done because these values are
         * interpreted as minutes. Please see table 8, SDS13781-4.
         */
        if (127 < duration)
        {
          duration = (duration - 127) * 60; /* minutes */
        }
        TimerStart(&SupervisionTimer, duration * 1000);
      }
    }
    else {
      // Duration == 0. I.e. the command has already been handled and completed. Clear the "More status updates coming" flag
      pArgs->properties1 &= (CC_SUPERVISION_ADD_MORE_STATUS_UPDATE(CC_SUPERVISION_MORE_STATUS_UPDATES_THIS_IS_LAST) | 0x7F);
    }
  }
  else
  {
    // Status for all other commands
    pArgs->properties1 &= (CC_SUPERVISION_ADD_MORE_STATUS_UPDATE(CC_SUPERVISION_MORE_STATUS_UPDATES_THIS_IS_LAST) | 0x7F);
    pArgs->duration = 0;
  }
}



static void EventHandlerApp(void)
{
  DPRINT("Got event!\r\n");

  uint8_t event;

  while (xQueueReceive(m_AppEventQueue, (uint8_t*)(&event), 0) == pdTRUE)
  {
    DPRINTF("Event: %d\r\n", event);
    AppStateManager((EVENT_APP)event);
  }
}

/**
 * @brief See description for function prototype in ZW_basis_api.h.
 */
ZW_APPLICATION_STATUS
ApplicationInit(EResetReason_t eResetReason)
{
  // NULL - We dont have the Application Task handle yet
  AppTimerInit(EAPPLICATIONEVENT_TIMER, NULL);

  g_eResetReason = eResetReason;

  /* hardware initialization */
  Board_Init();
  BRD420xBoardInit(RadioConfig.eRegion);

#ifdef DEBUGPRINT
  ZAF_UART0_enable(9600, true, false);
  DebugPrintConfig(m_aDebugPrintBuffer, sizeof(m_aDebugPrintBuffer), ZAF_UART0_tx_send);
#endif

  /* Init state machine*/
  currentState = STATE_APP_STARTUP;

  DPRINT("\n\n--------------------------------\n");
  DPRINT("Z-Wave Sample App: Switch On/Off\n");
  DPRINTF("SDK: %d.%d.%d ZAF: %d.%d.%d.%d [Freq: %d]\n",
          SDK_VERSION_MAJOR,
          SDK_VERSION_MINOR,
          SDK_VERSION_PATCH,
          ZAF_GetAppVersionMajor(),
          ZAF_GetAppVersionMinor(),
          ZAF_GetAppVersionPatchLevel(),
          ZAF_BUILD_NO,
          RadioConfig.eRegion);
  DPRINT("--------------------------------\n");
  DPRINTF("%s: Toggle switch on/off\n", Board_GetButtonLabel(SWITCH_TOGGLE_BTN));
  DPRINTF("%s: Toggle learn mode\n", Board_GetButtonLabel(APP_BUTTON_LEARN_RESET));
  DPRINT("      Hold 5 sec: Reset\n");
  DPRINTF("%s: Learn mode + identify\n", Board_GetLedLabel(APP_LED_INDICATOR));
  DPRINTF("%s: Switch status on/off\n", Board_GetLedLabel(SWITCH_STATUS_LED));
  DPRINT("--------------------------------\n\n");

  DPRINTF("ApplicationInit eResetReason = %d\n", g_eResetReason);

  CommandClassZWavePlusInfoInit(ZWavePlusEndpointIcons,
                                  sizeof(ZWavePlusEndpointIcons)/sizeof(ST_ENDPOINT_ICONS));







  CommandClassSupervisionInit(
      CC_SUPERVISION_STATUS_UPDATES_NOT_SUPPORTED,
      ZCB_CommandClassSupervisionGetReceived,
      NULL);

  CC_Version_SetApplicationVersionInfo(ZAF_GetAppVersionMajor(),
                                       ZAF_GetAppVersionMinor(),
                                       ZAF_GetAppVersionPatchLevel(),
                                       ZAF_BUILD_NO);

  Transport_AddEndpointSupport( &endPointFunctionality, endpointsNIF, NUMBER_OF_ENDPOINTS);

  /* Register task function */
  bool bWasTaskCreated = ZW_ApplicationRegisterTask(
                                                    ApplicationTask,
                                                    EAPPLICATIONEVENT_ZWRX,
                                                    EAPPLICATIONEVENT_ZWCOMMANDSTATUS,
                                                    &ProtocolConfig
                                                    );
  ASSERT(bWasTaskCreated);

  return(APPLICATION_RUNNING);
}

/**
 * A pointer to this function is passed to ZW_ApplicationRegisterTask() making it the FreeRTOS
 * application task.
 */
static void
ApplicationTask(SApplicationHandles* pAppHandles)
{
  // Init
  DPRINT("Enabling watchdog\n");
  WDOGn_Enable(DEFAULT_WDOG, true);

  g_AppTaskHandle = xTaskGetCurrentTaskHandle();
  g_pAppHandles = pAppHandles;
  AppTimerSetReceiverTask(g_AppTaskHandle);
  AppTimerRegister(&NotificationTimer, true, ZCB_NotificationTimerCallback);

  AppTimerRegister(&SupervisionTimer, false, ZCB_SupervisionTimerCallback);

  ZAF_Init(g_AppTaskHandle, pAppHandles, &ProtocolConfig, NULL);

  // Initialize Queue Notifier for events in the application.
  m_AppEventQueue = xQueueCreateStatic(
    sizeof_array(eventQueueStorage),
    sizeof(eventQueueStorage[0]),
    (uint8_t*)eventQueueStorage,
    &m_AppEventQueueObject
  );

  QueueNotifyingInit(
      &m_AppEventNotifyingQueue,
      m_AppEventQueue,
      g_AppTaskHandle,
      EAPPLICATIONEVENT_APP);

  ZAF_EventHelperInit(&m_AppEventNotifyingQueue);

  ZAF_JobHelperInit();
  ZAF_EventHelperEventEnqueue(EVENT_APP_INIT);







  Board_EnableButton(APP_BUTTON_LEARN_RESET);
  Board_EnableButton(SWITCH_TOGGLE_BTN);
  Board_EnableButton(SWITCH_TOGGLE_BTN2);

  Board_IndicatorInit(APP_LED_INDICATOR);
  Board_IndicateStatus(BOARD_STATUS_IDLE);

  EventDistributorConfig(
    &g_EventDistributor,
    sizeof_array(g_aEventHandlerTable),
    g_aEventHandlerTable,
    NULL
    );

  // Wait for and process events
  DPRINT("SwitchOnOff Event processor Started\r\n");
  uint32_t iMaxTaskSleep = 10;
  for (;;)
  {
    EventDistributorDistribute(&g_EventDistributor, iMaxTaskSleep, 0);
  }
}

/**
 * @brief See description for function prototype in ZW_TransportEndpoint.h.
 */
received_frame_status_t
Transport_ApplicationCommandHandlerEx(
  RECEIVE_OPTIONS_TYPE_EX *rxOpt,
  ZW_APPLICATION_TX_BUFFER *pCmd,
  uint8_t cmdLength)
{
  received_frame_status_t frame_status = RECEIVED_FRAME_STATUS_NO_SUPPORT;
  /* Call command class handlers */
  switch (pCmd->ZW_Common.cmdClass)
  {
    case COMMAND_CLASS_VERSION:
      frame_status = handleCommandClassVersion(rxOpt, pCmd, cmdLength);
      break;

    case COMMAND_CLASS_ASSOCIATION_GRP_INFO:
      frame_status = CC_AGI_handler( rxOpt, pCmd, cmdLength);
      break;

    case COMMAND_CLASS_ASSOCIATION:
      frame_status = handleCommandClassAssociation(rxOpt, pCmd, cmdLength);
      break;

    case COMMAND_CLASS_INDICATOR:
      frame_status = handleCommandClassIndicator(rxOpt, pCmd, cmdLength);
      break;

    case COMMAND_CLASS_POWERLEVEL:
      frame_status = handleCommandClassPowerLevel(rxOpt, pCmd, cmdLength);
      break;

    case COMMAND_CLASS_MANUFACTURER_SPECIFIC:
      frame_status = handleCommandClassManufacturerSpecific(rxOpt, pCmd, cmdLength);
      break;

    case COMMAND_CLASS_ZWAVEPLUS_INFO:
      frame_status = handleCommandClassZWavePlusInfo(rxOpt, pCmd, cmdLength);
      break;

    case COMMAND_CLASS_BASIC:
      frame_status = CC_Basic_handler(rxOpt, pCmd, cmdLength);
      break;

    case COMMAND_CLASS_SWITCH_BINARY_V2:
        frame_status = handleCommandClassBinarySwitch(rxOpt, pCmd, cmdLength);
        break;

      case COMMAND_CLASS_NOTIFICATION_V3:
        frame_status = handleCommandClassNotification(rxOpt, pCmd, cmdLength);
        break;





      case COMMAND_CLASS_SUPERVISION:
        frame_status = handleCommandClassSupervision(rxOpt, pCmd, cmdLength);
        break;

      case COMMAND_CLASS_MULTI_CHANNEL_ASSOCIATION_V2:
        frame_status = handleCommandClassMultiChannelAssociation(rxOpt, pCmd, cmdLength);
        break;
      /*case COMMAND_CLASS_SECURITY:
      case COMMAND_CLASS_SECURITY_2:
        frame_status = handleCommandClassSecurity(rxOpt, pCmd, cmdLength);
        break;*/

      case COMMAND_CLASS_FIRMWARE_UPDATE_MD_V5:
        frame_status = handleCommandClassFWUpdate(rxOpt, pCmd, cmdLength);
      break;

      //case COMMAND_CLASS_MULTI_CHANNEL_V4:
        //frame_status = MultiChanCommandHandler(rxOpt, pCmd, cmdLength);
        //break;
  }
  return frame_status;
}

/**
 * @brief See description for function prototype in misc.h.
 */
uint8_t
GetMyNodeID(void)
{
  return g_pAppHandles->pNetworkInfo->NodeId;
}

static void
LearnCompleted(void)
{
  uint8_t bNodeID = GetMyNodeID();

  /*If bNodeID= 0xff.. learn mode failed*/
  if(NODE_BROADCAST != bNodeID)
  {
	  /* Success */
	      if (0 == bNodeID)
	      {
	        SetDefaultConfiguration();
	        if(ESWTIMER_STATUS_FAILED != notificationOverLoadTimerStatus)
	        {
	          TimerStop(&NotificationTimer);
	          notificationOverLoadTimerStatus = ESWTIMER_STATUS_FAILED;
	        }
	      }
  }
  ZAF_EventHelperEventEnqueue((EVENT_APP) EVENT_SYSTEM_LEARNMODE_FINISHED);
  Transport_OnLearnCompleted(bNodeID);
}

/**
 * @brief Returns the current state of the application state machine.
 * @return Current state
 */
STATE_APP
GetAppState(void)
{
  return currentState;
}

/**
 * @brief The core state machine of this sample application.
 * @param event The event that triggered the call of AppStateManager.
 */
void
AppStateManager(EVENT_APP event)
{
  DPRINTF("AppStateManager St: %d, Ev: %d\r\n", currentState, event);

  if ((BTN_EVENT_LONG_PRESS(APP_BUTTON_LEARN_RESET) == (BUTTON_EVENT)event) ||
      (EVENT_SYSTEM_RESET == (EVENT_SYSTEM)event))
  {
    /*Force state change to activate system-reset without taking care of current state.*/
    ChangeState(STATE_APP_RESET);
    /* Send reset notification*/
    DeviceResetLocally();
  }

  switch(currentState)
  {
    case STATE_APP_STARTUP:

      if(EVENT_APP_FLUSHMEM_READY == event)
      {
        AppResetNvm();
      }

      if(EVENT_APP_INIT == event)
      {

        /* Load the application settings from NVM3 file system */
        LoadConfiguration();
        /*it is not the time to wake completely*/

        /* Initialize association module */
        AssociationInitEndpointSupport(false,
                                           rootGroupMapping,
                                           sizeof(rootGroupMapping)/sizeof(ASSOCIATION_ROOT_GROUP_MAPPING),
                                           pFileSystemApplication);

        // Setup AGI group lists
        AGI_Init();
        // Root device
        CC_AGI_LifeLineGroupSetup(agiTableLifeLine,
        		                  sizeof_array(agiTableLifeLine),
								  ENDPOINT_ROOT );
        // Endpoint 1
        CC_AGI_LifeLineGroupSetup(agiTableLifeLineEP1,
                                  sizeof_array(agiTableLifeLineEP1),
                                  ENDPOINT_1);
        // Endpoint 2
        CC_AGI_LifeLineGroupSetup(agiTableLifeLineEP2,
                                  sizeof_array(agiTableLifeLineEP2),
                                  ENDPOINT_2);

        AGI_ResourceGroupSetup(agiTableRootDeviceGroups,
                               sizeof_array(agiTableRootDeviceGroups),
                               ENDPOINT_ROOT);
        // Endpoint 1
        AGI_ResourceGroupSetup(agiTableEndpoint1Groups,
                               sizeof_array(agiTableEndpoint1Groups),
                               ENDPOINT_1);

        // Endpoint 2
        AGI_ResourceGroupSetup(agiTableEndpoint2Groups,
                               sizeof_array(agiTableEndpoint2Groups),
                               ENDPOINT_2);



        /*
         * Initialize Event Scheduler.
         */
        Transport_OnApplicationInitSW( &m_AppNIF, NULL);

        /* Enter SmartStart*/
        /* Protocol will commence SmartStart only if the node is NOT already included in the network */
        ZAF_setNetworkLearnMode(E_NETWORK_LEARN_MODE_INCLUSION_SMARTSTART, g_eResetReason);

        /* Init state machine*/
        ZAF_EventHelperEventEnqueue(EVENT_EMPTY);

        break;
      }
      OtaInit(NULL,NULL,NULL);
      ChangeState(STATE_APP_IDLE);
      break;

    case STATE_APP_IDLE:

      if (EVENT_APP_REFRESH_MMI == event)
      {
        Board_IndicateStatus(BOARD_STATUS_IDLE);
        RefreshMMI();
      }

      if(EVENT_APP_FLUSHMEM_READY == event)
      {
        AppResetNvm();
        LoadConfiguration();
      }

      if(EVENT_APP_SMARTSTART_IN_PROGRESS == event)
      {
        ChangeState(STATE_APP_LEARN_MODE);
      }

      if ((BTN_EVENT_SHORT_PRESS(APP_BUTTON_LEARN_RESET) == (BUTTON_EVENT)event) ||
          (EVENT_SYSTEM_LEARNMODE_START == (EVENT_SYSTEM)event))
      {
        DPRINT("APP_BUTTON_LEARN_RESET SHORT_PRESS\n");

        if (EINCLUSIONSTATE_EXCLUDED != g_pAppHandles->pNetworkInfo->eInclusionState)
        {
          DPRINT("LEARN_MODE_EXCLUSION\r\n");
          ZAF_setNetworkLearnMode(E_NETWORK_LEARN_MODE_EXCLUSION_NWE, g_eResetReason);
        }
        else
        {
          DPRINT("LEARN_MODE_INCLUSION\r\n");
          ZAF_setNetworkLearnMode(E_NETWORK_LEARN_MODE_INCLUSION, g_eResetReason);
        }
        ChangeState(STATE_APP_LEARN_MODE);
      }

      if (BTN_EVENT_SHORT_PRESS(SWITCH_TOGGLE_BTN) == (BUTTON_EVENT)event)
      {
        DPRINT("BIN_SWITCH_1 ");
        //ToggleSwitch(BIN_SWITCH_1);
        /* Update the lifeline destinations when the Binary Switch state has changed */
        void * pData = CC_BinarySwitch_prepare_zaf_tse_data(&zaf_tse_local_actuation);
        ZAF_TSE_Trigger((void *)CC_BinarySwitch_report_stx, pData, true);
      }

      if (BTN_EVENT_SHORT_PRESS(SWITCH_TOGGLE_BTN2) == (BUTTON_EVENT)event)
       {
        DPRINT("BIN_SWITCH_2 ");
        //ToggleSwitch(BIN_SWITCH_2);
        /* Update the lifeline destinations when the Binary Switch state has changed */
        void * pData = CC_BinarySwitch_prepare_zaf_tse_data(&zaf_tse_local_actuation2);
        ZAF_TSE_Trigger((void *)CC_BinarySwitch_report_stx, pData, true);
       }


      break;

    case STATE_APP_LEARN_MODE:
      if(EVENT_APP_REFRESH_MMI == event)
      {
        Board_IndicateStatus(BOARD_STATUS_LEARNMODE_ACTIVE);
      }

      if(EVENT_APP_FLUSHMEM_READY == event)
      {
        AppResetNvm();
        LoadConfiguration();
      }

      if ((BTN_EVENT_SHORT_PRESS(APP_BUTTON_LEARN_RESET) == (BUTTON_EVENT)event) ||
          (EVENT_SYSTEM_LEARNMODE_STOP == (EVENT_SYSTEM)event))
      {
        ZAF_setNetworkLearnMode(E_NETWORK_LEARN_MODE_DISABLE, g_eResetReason);

        //Go back to smart start if the node was never included.
        //Protocol will not commence SmartStart if the node is already included in the network.
        ZAF_setNetworkLearnMode(E_NETWORK_LEARN_MODE_INCLUSION_SMARTSTART, g_eResetReason);

        Board_IndicateStatus(BOARD_STATUS_IDLE);
        ChangeState(STATE_APP_IDLE);

        /* If we are in a network and the Identify LED state was changed to idle due to learn mode, report it to lifeline */
        CC_Indicator_RefreshIndicatorProperties();
        ZAF_TSE_Trigger((void *)CC_Indicator_report_stx, &ZAF_TSE_localActuationIdentifyData, true);
      }

      if(EVENT_SYSTEM_LEARNMODE_FINISHED == (EVENT_SYSTEM)event)
      {
        //Go back to smart start if the node was excluded.
        //Protocol will not commence SmartStart if the node is already included in the network.
        ZAF_setNetworkLearnMode(E_NETWORK_LEARN_MODE_INCLUSION_SMARTSTART, g_eResetReason);

        ChangeState(STATE_APP_IDLE);

        /* If we are in a network and the Identify LED state was changed to idle due to learn mode, report it to lifeline */
        CC_Indicator_RefreshIndicatorProperties();
        ZAF_TSE_Trigger((void *)CC_Indicator_report_stx, &ZAF_TSE_localActuationIdentifyData, true);
      }
      break;

    case STATE_APP_RESET:
      if(EVENT_APP_REFRESH_MMI == event){}

      if(EVENT_APP_FLUSHMEM_READY == event)
      {
        AppResetNvm();
        /* Soft reset */
        Board_ResetHandler();
      }
      break;


    case STATE_APP_TRANSMIT_DATA:

          if(EVENT_APP_FLUSHMEM_READY == event)
          {
            AppResetNvm();
            LoadConfiguration();
          }

          if (EVENT_APP_FINISH_EVENT_JOB == event)
          {
            ClearLastNotificationAction(LIFELANE_PROFILE, notificationOverLoadendpoint);
            ChangeState(STATE_APP_IDLE);
          }


          break;

    default:
      // Do nothing.
      break;
  }
}

/**
 * @brief Sets the current state to a new, given state.
 * @param newState New state.
 */
static void
ChangeState(STATE_APP newState)
{
  currentState = newState;
  /**< Pre-action on new state is to refresh MMI */
  ZAF_EventHelperEventEnqueue(EVENT_APP_REFRESH_MMI);
}

/**
 * @brief Transmission callback for Device Reset Locally call.
 * @param pTransmissionResult Result of each transmission.
 */
void
DeviceResetLocallyDone(TRANSMISSION_RESULT * pTransmissionResult)
{
  if (TRANSMISSION_RESULT_FINISHED == pTransmissionResult->isFinished)
  {
    /* Reset protocol */
    // Set default command to protocol
    SZwaveCommandPackage CommandPackage;
    CommandPackage.eCommandType = EZWAVECOMMANDTYPE_SET_DEFAULT;

    DPRINT("\nDisabling watchdog during reset\n");
    WDOGn_Enable(DEFAULT_WDOG, false);

    EQueueNotifyingStatus Status = QueueNotifyingSendToBack(g_pAppHandles->pZwCommandQueue, (uint8_t*)&CommandPackage, 500);
    ASSERT(EQUEUENOTIFYING_STATUS_SUCCESS == Status);
  }
}

/**
 * @brief Send reset notification.
 */
void
DeviceResetLocally(void)
{
  AGI_PROFILE lifelineProfile = {
		  ASSOCIATION_GROUP_INFO_REPORT_PROFILE_GENERAL_V2,
		   ASSOCIATION_GROUP_INFO_REPORT_PROFILE_GENERAL_LIFELINE_V2
  };
  DPRINT("Call locally reset\r\n");

  CC_DeviceResetLocally_notification_tx(&lifelineProfile, DeviceResetLocallyDone);
}

/**
 * @brief See description for function prototype in CC_Version.h.
 */
uint8_t
CC_Version_getNumberOfFirmwareTargets_handler(void)
{
  return 1; /*CHANGE THIS - firmware 0 version*/
}

/**
 * @brief See description for function prototype in CC_Version.h.
 */
void
handleGetFirmwareVersion(
  uint8_t bFirmwareNumber,
  VG_VERSION_REPORT_V2_VG *pVariantgroup)
{
  /*firmware 0 version and sub version*/
  if(bFirmwareNumber == 0)
  {
    pVariantgroup->firmwareVersion = ZAF_GetAppVersionMajor();
    pVariantgroup->firmwareSubVersion = ZAF_GetAppVersionMinor();
  }
  else
  {
    /*Just set it to 0 if firmware n is not present*/
    pVariantgroup->firmwareVersion = 0;
    pVariantgroup->firmwareSubVersion = 0;
  }
}

/**
 * Report current switch value to binary switch command class handler
 *
 * @brief See description for function prototype in CC_BinarySwitch.h
 */
CMD_CLASS_BIN_SW_VAL
appBinarySwitchGetCurrentValue(uint8_t endpoint)
{
	if(endpoint==ENDPOINT_1){
	  if (BIN_SWITCH1.switchStatus & SWITCH_ON)
	      {
	        return CMD_CLASS_BIN_ON;
	      }
	      else
	      {
	        return CMD_CLASS_BIN_OFF;
	      }
  }else{
	  if (BIN_SWITCH2.switchStatus & SWITCH_ON)
	      {
	        return CMD_CLASS_BIN_ON;
	      }
	      else
	      {
	        return CMD_CLASS_BIN_OFF;
	      }

  }
}

/**
 * Report current target value to binary switch command class handler
 *
 * @brief See description for function prototype in CC_BinarySwitch.h
 */
CMD_CLASS_BIN_SW_VAL
appBinarySwitchGetTargetValue(uint8_t endpoint)
{
  /* Always report target value equal to current value  */
  return appBinarySwitchGetCurrentValue(endpoint);
}

/**
 * Report current duration to binary switch command class handler
 *
 * @brief See description for function prototype in CC_BinarySwitch.h
 */
uint8_t
appBinarySwitchGetDuration(uint8_t endpoint)
{
  UNUSED(endpoint);

  /* Always report instant transition; the only reasonable transition for a binary switch. */
  return 0;
}

/**
 * Report current duration to binary switch command class handler
 *
 * @brief See description for function prototype in CC_BinarySwitch.h
 */
e_cmd_handler_return_code_t
appBinarySwitchSet(
  CMD_CLASS_BIN_SW_VAL val,
  uint8_t duration,
  uint8_t endpoint
)
{

  UNUSED(duration); /* Ignore duration - for a binary switch only instant transitions make sense */

  DPRINTF("\r\nappBinarySwitchSet %u\r\n", val);

  if(endpoint==ENDPOINT_1){
   if(CMD_CLASS_BIN_OFF == val)
    {

	  BIN_SWITCH1.switchStatus &= ~SWITCH_ON;
     // Board_SetLed(OUTLET1_STATUS_LED, LED_OFF);
    }
    else if(CMD_CLASS_BIN_ON == val)
    {
    	BIN_SWITCH1.switchStatus |= SWITCH_ON;
      //Board_SetLed(OUTLET1_STATUS_LED, LED_ON);
    }


  }else{
	  if(CMD_CLASS_BIN_OFF == val)
	      {

	  	  BIN_SWITCH2.switchStatus &= ~SWITCH_ON;
	       // Board_SetLed(OUTLET2_STATUS_LED, LED_OFF);
	      }
	      else if(CMD_CLASS_BIN_ON == val)
	      {
	      	BIN_SWITCH2.switchStatus |= SWITCH_ON;
	        //Board_SetLed(OUTLET2_STATUS_LED, LED_ON);
	      }


  }

    return E_CMD_HANDLER_RETURN_CODE_HANDLED;





}

/**
 * Report factory default switch transition duration to binary switch command class handler
 *
 * @brief See description for function prototype in CC_BinarySwitch.h
 */
uint8_t
appBinarySwitchGetFactoryDefaultDuration(uint8_t endpoint)
{
  UNUSED(endpoint);
  return 0;
}

/**
 * @brief Function resets configuration to default values.
 */
void
SetDefaultConfiguration(void)
{
  AssociationInit(true, pFileSystemApplication);
  BIN_SWITCH1.switchStatus = CMD_CLASS_BIN_OFF;
  BIN_SWITCH2.switchStatus = CMD_CLASS_BIN_OFF;

    SNotificationData tNotificationData;
     memset(&tNotificationData, 1, sizeof(SNotificationData));

     Ecode_t errCode = nvm3_writeData(pFileSystemApplication, ZAF_FILE_ID_NOTIFICATIONDATA, &tNotificationData, sizeof(SNotificationData));
     ASSERT(ECODE_NVM3_OK == errCode); //Assert has been kept for debugging , can be removed from production code. This error can only be caused by some internal flash HW failure

     uint32_t appVersion = ZAF_GetAppVersion();
     errCode = nvm3_writeData(pFileSystemApplication, ZAF_FILE_ID_APP_VERSION, &appVersion, ZAF_FILE_SIZE_APP_VERSION);
     ASSERT(ECODE_NVM3_OK == errCode); //Assert has been kept for debugging , can be removed from production code if this error can only be caused by some internal flash HW failure
}

/**
 * @brief This function loads the application settings from file system.
 * If no settings are found, default values are used and saved.
 */
bool
LoadConfiguration(void)
{
  // Init file system
  ApplicationFileSystemInit(&pFileSystemApplication);

  /* Check to see, if any valid configuration is stored in the file system */
  SFileSet FileSet = {
    .pFileSystem = pFileSystemApplication,
    .iFileCount = sizeof_array(g_aFileDescriptors),
    .pFileDescriptors = g_aFileDescriptors
  };

  EFileSysVerifyStatus eFileSetVerifyStatus = FileSystemVerifySet(&FileSet, g_aFileStatus);
  if (EFSV_STATUS_SUCCESS == eFileSetVerifyStatus)
  {

    ApplicationData = readAppData();
    onOffState = ApplicationData.onOffState;

    /* Initialize association module */
    AssociationInit(false, pFileSystemApplication);
    /* Initialize PowerLevel functionality*/
        loadStatusPowerLevel();

    return true;
  }
  else
  {
    DPRINT("Application FileSystem Verify failed\r\n");
    loadInitStatusPowerLevel();

    uint32_t objectType;
    size_t   dataLen;
    Ecode_t  versionFileStatus = nvm3_findObject(pFileSystemApplication, ZAF_FILE_ID_APP_VERSION, &objectType, &dataLen);

    if(ECODE_NVM3_OK != versionFileStatus)
    {
      // Reset the file system if ZAF_FILE_ID_APP_VERSION is missing since this indicates
      // corrupt or missing file system.
      AppResetNvm();
    }
    else
    {
      // Add code for migration of file system to higher version here.
      // Use info in g_aFileStatus[] to see what files are missing or
      // have changed size.
    }

    return false;
  }
}

void AppResetNvm(void)
{
  DPRINT("Resetting application FileSystem to default\r\n");

  ASSERT(0 != pFileSystemApplication); //Assert has been kept for debugging , can be removed from production code. This error can only be caused by some internal flash HW failure

  Ecode_t errCode = nvm3_eraseAll(pFileSystemApplication);
  ASSERT(ECODE_NVM3_OK == errCode); //Assert has been kept for debugging , can be removed from production code. This error can only be caused by some internal flash HW failure

  /* Initialize transport layer */
  Transport_SetDefault();

  /* Apparently there is no valid configuration in file system, so load */
  /* default values and save them to file system. */
  SetDefaultConfiguration();
}

/**
 * @brief Handler for basic set. Handles received basic set commands.
 * @param[in] val Parameter dependent of the application device class.
 * @param[in] endpoint Endpoint
 * @return command handler return code
 */
e_cmd_handler_return_code_t CC_Basic_Set_handler(uint8_t val, uint8_t endpoint)
{
  /* Basic Set Command is mapped to Binary Switch Set */
  return CommandClassBinarySwitchSupportSet(val, 0, endpoint);
}

/**
 * @brief Handler for basic get. Handles received basic get commands.
 */
uint8_t
CC_Basic_GetCurrentValue_handler(uint8_t endpoint)
{
  return appBinarySwitchGetCurrentValue(endpoint);
}

/**
 * @brief Report the target value
 * @return target value.
 */
uint8_t
CC_Basic_GetTargetValue_handler(uint8_t endpoint)
{
  return appBinarySwitchGetTargetValue(endpoint);
}

/**
 * @brief Report transition duration time.
 * @return duration time.
 */
uint8_t
CC_Basic_GetDuration_handler(uint8_t endpoint)
{
  return appBinarySwitchGetDuration(endpoint);
}

/**
 * @brief Toggles LED state variable and refreshes MMI.
 */
void
ToggleLed(void)
{
  if (CMD_CLASS_BIN_ON == onOffState)
  {
    onOffState = CMD_CLASS_BIN_OFF;
  }
  else
  {
    onOffState = CMD_CLASS_BIN_ON;
  }
  RefreshMMI();
  ApplicationData.onOffState = onOffState;
  writeAppData(&ApplicationData);
}

/**
 * @brief Refreshes MMI.
 */
void
RefreshMMI(void)
{
  if (CMD_CLASS_BIN_OFF == onOffState)
  {
    Board_SetLed(SWITCH_STATUS_LED, LED_OFF);
    DPRINT("Binary Switch OFF\r\n");
  }
  else if (CMD_CLASS_BIN_ON == onOffState)
  {
    Board_SetLed(SWITCH_STATUS_LED, LED_ON);
    DPRINT("Binary Switch ON\r\n");
  }
}

/**
 * @brief Reads application data from file system.
 */
SApplicationData
readAppData(void)
{
  ASSERT(0 != pFileSystemApplication); //Assert has been kept for debugging , can be removed from production code. This error can only be caused by some internal flash HW failure

  SApplicationData AppData;
  Ecode_t errCode = nvm3_readData(pFileSystemApplication, FILE_ID_APPLICATIONDATA, &AppData, sizeof(SApplicationData));
  ASSERT(ECODE_NVM3_OK == errCode); //Assert has been kept for debugging , can be removed from production code. This error hard to occur when a corresponing write is successfull
                                    //Can only happended in case of some hardware failure

  return AppData;
}

/**
 * @brief Writes application data to file system.
 */
void writeAppData(const SApplicationData* pAppData)
{
  ASSERT(0 != pFileSystemApplication); //Assert has been kept for debugging , can be removed from production code. This means nvm3_open failed can only be caused by some internal flash HW failure

  Ecode_t errCode = nvm3_writeData(pFileSystemApplication, FILE_ID_APPLICATIONDATA, pAppData, sizeof(SApplicationData));
  ASSERT(ECODE_NVM3_OK == errCode); //Assert has been kept for debugging , can be removed from production code. This error can only be caused by some internal flash HW failure
}

uint16_t handleFirmWareIdGet(uint8_t n)
{
  if (n == 0)
  {
    // Firmware 0
    return APP_FIRMWARE_ID;
  }
  // Invalid Firmware number.
  return 0;
}


/**
 * This function is used to return a pointer for input data for the TSE trigger.
 * @param rxOption pointer that will be included in the ZAF_TSE input data.
*/
void* CC_Basic_prepare_zaf_tse_data(RECEIVE_OPTIONS_TYPE_EX* pRxOpt)
{
  /* Basic is mapped to Binary Switch for Root and EP1
  Basic is mapped to multilevel Switch for EP3.
  Call here different functions
  for each endpoint (rxOpt->destNode.endpoint) if they do not map to the same
  Command Class & Command */
  if (0 == pRxOpt->destNode.endpoint || 1 == pRxOpt->destNode.endpoint)
  {
    return CC_BinarySwitch_prepare_zaf_tse_data(pRxOpt);
  }
  else if (2 == pRxOpt->destNode.endpoint)
  {
    return CC_BinarySwitch_prepare_zaf_tse_data(pRxOpt);
  }
  else
  {
    /* Basic is not supported, return NULL so the TSE trigger will just be rejected */
    return NULL;
  }
}






/**
 * This function is used to return a pointer for input data for the TSE trigger.
 * @param rxOption pointer that will be included in the ZAF_TSE input data.
*/

void CC_Basic_report_stx(TRANSMIT_OPTIONS_TYPE_SINGLE_EX txOptions, void* pData)
{
  /* Basic Report is mapped to different commands for all end points.
  Call here different functions for each endpoint (pData->rxOpt->destNode.endpoint)
  if they do not map to the same Command Class & Command */
  s_zaf_tse_data_input_template_t* pDataInput = (s_zaf_tse_data_input_template_t*)(pData);
  RECEIVE_OPTIONS_TYPE_EX RxOptions = pDataInput->rxOptions;

  if (0 == RxOptions.destNode.endpoint || 1 == RxOptions.destNode.endpoint)
  {
    CC_BinarySwitch_report_stx(txOptions,pData);
  }
  else if (2 == RxOptions.destNode.endpoint)
  {
	  CC_BinarySwitch_report_stx(txOptions,pData);
  }
}


/**
 * Considering several endpoints supporting the same Command Class,
 * this function finds the index of the end point if they were stored in an array.
 * e.g. if Endpoints 0, 3, 5 and 12 support the Command Class, their index would
 * be respectively 0, 1, 2 and 3.
 * @param endpoint Indicating the endpoint identifier to search for.
*/
uint8_t CC_BinarySwitch_getEndpointIndex(uint8_t endpoint)
{
  for (uint8_t i=0; i<sizeof(endpoints_Supporting_BinarySwitch_CC); i++)
  {
    if (endpoint == endpoints_Supporting_BinarySwitch_CC[i]){
      return i;
    }
  }
  /*End point was not found, this should not happen but in this case, just return the first End Point */
  return 0;
}



/**
 * This function is used to notify the Application that the CC Basic Set
 * status is in a WORKING state. The application can subsequently make the TSE Trigger
 * using the information passed on from the rxOptions.
 * @param rxOption pointer used to when triggering the "working state"
*/
void CC_Basic_report_notifyWorking(RECEIVE_OPTIONS_TYPE_EX* pRxOpt)
{
  /* Handle here the case where Basic CC has triggered a working state.
  Look in the Door Lock keypad application for an example. */
  UNUSED(pRxOpt);
}












/**
 * Prepare the data input for the TSE for any Binary Switch CC command based on the pRxOption pointer.
 * @param pRxOpt pointer used to indicate how the frame was received (if any) and what endpoints are affected
*/


void* CC_BinarySwitch_prepare_zaf_tse_data(RECEIVE_OPTIONS_TYPE_EX* pRxOpt)
{
  uint8_t endpoint_index = CC_BinarySwitch_getEndpointIndex(pRxOpt->destNode.endpoint);
  memset(&ZAF_TSE_BinarySwitchData[endpoint_index], 0, sizeof(s_CC_binarySwitch_data_t));
  ZAF_TSE_BinarySwitchData[endpoint_index].rxOptions = *pRxOpt;

  /* Apply here a endpoint mapping, endpoint 0 state maps to endpoint 1.
  Note: in order to map the root device to more than one end point,
  it is recommended to modify the Command handler to be called for each mapped End point
  e.g. call CommandClassBinarySwitchSupportSet() for every mapped
  end point in handleCommandClassBinarySwitch() */
  if (0 == ZAF_TSE_BinarySwitchData[endpoint_index].rxOptions.destNode.endpoint)
  {
    ZAF_TSE_BinarySwitchData[endpoint_index].rxOptions.destNode.endpoint = 1;
  }

  if (1 == ZAF_TSE_BinarySwitchData[endpoint_index].rxOptions.destNode.endpoint)
    {
      ZAF_TSE_BinarySwitchData[endpoint_index].rxOptions.destNode.endpoint = 2;
    }

  return &ZAF_TSE_BinarySwitchData[endpoint_index];
}














uint8_t CC_Version_GetHardwareVersion_handler(void)
{
  return 1;
}

void CC_ManufacturerSpecific_ManufacturerSpecificGet_handler(uint16_t * pManufacturerID,
                                                             uint16_t * pProductID)
{
  *pManufacturerID = APP_MANUFACTURER_ID;
  *pProductID = APP_PRODUCT_ID;
}

/*
 * This function will report a serial number in a binary format according to the specification.
 * The serial number is the chip serial number and can be verified using the Simplicity Commander.
 * The choice of reporting can be changed in accordance with the Manufacturer Specific
 * Command Class specification.
 */
void CC_ManufacturerSpecific_DeviceSpecificGet_handler(device_id_type_t * pDeviceIDType,
                                                       device_id_format_t * pDeviceIDDataFormat,
                                                       uint8_t * pDeviceIDDataLength,
                                                       uint8_t * pDeviceIDData)
{
  *pDeviceIDType = DEVICE_ID_TYPE_SERIAL_NUMBER;
  *pDeviceIDDataFormat = DEVICE_ID_FORMAT_BINARY;
  *pDeviceIDDataLength = 8;
  uint64_t uuID = SYSTEM_GetUnique();
  DPRINTF("\r\nuuID: %x", (uint32_t)uuID);
  *(pDeviceIDData + 0) = (uint8_t)(uuID >> 56);
  *(pDeviceIDData + 1) = (uint8_t)(uuID >> 48);
  *(pDeviceIDData + 2) = (uint8_t)(uuID >> 40);
  *(pDeviceIDData + 3) = (uint8_t)(uuID >> 32);
  *(pDeviceIDData + 4) = (uint8_t)(uuID >> 24);
  *(pDeviceIDData + 5) = (uint8_t)(uuID >> 16);
  *(pDeviceIDData + 6) = (uint8_t)(uuID >>  8);
  *(pDeviceIDData + 7) = (uint8_t)(uuID >>  0);
}

void
ZCB_JobStatus(TRANSMISSION_RESULT * pTransmissionResult)
{
  DPRINTF("\r\nTX CB for N %u", pTransmissionResult->nodeId);

  if (TRANSMISSION_RESULT_FINISHED == pTransmissionResult->isFinished)
  {
    ZAF_EventHelperEventEnqueue(EVENT_APP_FINISH_EVENT_JOB);
  }
}


/**
 * @brief Called every time the notification timer triggers.
 */
void
ZCB_NotificationTimerCallback(SSwTimer *pTimer)
{
  UNUSED(pTimer);

  JOB_STATUS jobStatus;

  DPRINT("\r\nNtfctn timer");

  if (false == notificationOverLoadActiveState)
  {
    /*Find new endpoint to send notification*/
    notificationOverLoadendpoint++;
    if (notificationOverLoadendpoint > NUMBER_OF_ENDPOINTS)
    {
      notificationOverLoadendpoint = 1;
    }

    DPRINTF("\r\nNtfctn enable EP %u", notificationOverLoadendpoint);

    notificationOverLoadActiveState = true;
    NotificationEventTrigger(
        LIFELANE_PROFILE,
        NOTIFICATION_TYPE_POWER_MANAGEMENT,
        NOTIFICATION_EVENT_POWER_MANAGEMENT_OVERLOADED_DETECTED,
        NULL,
        0,
        notificationOverLoadendpoint);
  }
  else
  {
    DPRINTF("\r\nNtfctn disable EP %u", notificationOverLoadendpoint);

    notificationOverLoadActiveState = false;
    NotificationEventTrigger(
        LIFELANE_PROFILE,
        NOTIFICATION_TYPE_POWER_MANAGEMENT,
        NOTIFICATION_EVENT_POWER_MANAGEMENT_NO_EVENT,
        &supportedEvents,
        1,
        notificationOverLoadendpoint);
  }

  //@ [NOTIFICATION_TRANSMIT]
  jobStatus = UnsolicitedNotificationAction(
      LIFELANE_PROFILE,
      notificationOverLoadendpoint,
      ZCB_JobStatus);
  //@ [NOTIFICATION_TRANSMIT]

  if (JOB_STATUS_SUCCESS != jobStatus)
  {
    TRANSMISSION_RESULT transmissionResult;

    DPRINTF("\r\nX%u", jobStatus);

    transmissionResult.status = false;
    transmissionResult.nodeId = 0;
    transmissionResult.isFinished = TRANSMISSION_RESULT_FINISHED;

    ZCB_JobStatus(&transmissionResult);
  }

  ChangeState(STATE_APP_TRANSMIT_DATA);
}

/**
 * Transmits a Supervision report.
 *
 * This function is triggered by the timer set in the Supervision Get handler.
 */
void ZCB_SupervisionTimerCallback(SSwTimer *pTimer)
{
  TRANSMIT_OPTIONS_TYPE_SINGLE_EX * pTxOptions;
  RxToTxOptions(&rxOptionSupervision, &pTxOptions);
  CmdClassSupervisionReportSend(
      pTxOptions,
      CC_SUPERVISION_ADD_MORE_STATUS_UPDATE(CC_SUPERVISION_MORE_STATUS_UPDATES_THIS_IS_LAST) | CC_SUPERVISION_ADD_SESSION_ID(sessionID),
      CC_SUPERVISION_STATUS_SUCCESS,
      0);
}

